#ifndef CONDITION_H_
#define CONDITION_H_

#include <common/Common.h>
#include <common/toolkit/TimeTk.h>
#include "Mutex.h"


#include <linux/sched.h>


struct Condition;
typedef struct Condition Condition;

enum cond_wait_res;
typedef enum cond_wait_res cond_wait_res_t;

static inline void Condition_init(Condition* this);
static inline Condition* Condition_construct(void);
static inline void Condition_uninit(Condition* this);
static inline void Condition_destruct(Condition* this);

static inline cond_wait_res_t Condition_wait(Condition* this, Mutex* mutex);
static inline cond_wait_res_t Condition_timedwait(Condition* this,
   Mutex* mutex, int timeoutMS);
static inline void Condition_broadcast(Condition* this);
static inline void Condition_signal(Condition* this);


struct Condition
{
   wait_queue_head_t queue;
};

enum cond_wait_res
{
   COND_WAIT_TIMEOUT =  0,
   COND_WAIT_SUCCESS =  1,
   COND_WAIT_SIGNAL  = -1
};


void Condition_init(Condition* this)
{
   init_waitqueue_head(&this->queue);
}

Condition* Condition_construct(void)
{
   Condition* this = (Condition*)kmalloc(sizeof(Condition), GFP_KERNEL);
   
   Condition_init(this);
   
   return this;
}

void Condition_uninit(Condition* this)
{
}

void Condition_destruct(Condition* this)
{
   Condition_uninit(this);
   
   kfree(this);
}

cond_wait_res_t Condition_wait(Condition* this, Mutex* mutex)
{
   cond_wait_res_t retVal;
   wait_queue_t wait;
   
   init_waitqueue_entry(&wait, current);
   add_wait_queue(&this->queue, &wait);
   
   set_current_state(TASK_INTERRUPTIBLE);
   
   Mutex_unlock(mutex);
   schedule();
   retVal = signal_pending(current) ? COND_WAIT_SIGNAL : COND_WAIT_SUCCESS;
   
   __set_current_state(TASK_RUNNING);
   
   remove_wait_queue(&this->queue, &wait);
   
   Mutex_lock(mutex);
   
   return retVal;
}

cond_wait_res_t Condition_timedwait(Condition* this, Mutex* mutex, int timeoutMS)
{
   cond_wait_res_t retVal;
   long __timeout;

   wait_queue_t wait;
   init_waitqueue_entry(&wait, current);
   add_wait_queue(&this->queue, &wait);
   
   set_current_state(TASK_INTERRUPTIBLE);
   
   Mutex_unlock(mutex);

   // timeval to jiffies
   __timeout = TimeTk_msToJiffiesSchedulable(timeoutMS);  
   
   // wait
   if( (__timeout = schedule_timeout(__timeout) ) )
      retVal = signal_pending(current) ? COND_WAIT_SIGNAL : COND_WAIT_SUCCESS;
   else
      retVal = COND_WAIT_TIMEOUT;
   
   __set_current_state(TASK_RUNNING);
   
   remove_wait_queue(&this->queue, &wait);
   
   Mutex_lock(mutex);
   
   return retVal;
}

void Condition_broadcast(Condition* this)
{
   wake_up_interruptible(&this->queue);
}


/**
 * Note: this routine simply delegates to broadcast(..) for now
 */
void Condition_signal(Condition* this)
{
   Condition_broadcast(this);
}

#endif /*CONDITION_H_*/
